@@ -1,28 +1,27 @@
 from dataclasses import field
 from typing import Any, Dict, Optional
 
+import equinox.internal as eqxi
 import jax.numpy as jnp
 
 from .custom_types import Array, Bool, PyTree, Scalar
 from .global_interpolation import DenseInterpolation
-from .misc import ContainerMeta
 from .path import AbstractPath
 
 
-class RESULTS(metaclass=ContainerMeta):
+class RESULTS(metaclass=eqxi.ContainerMeta):
     successful = ""
+    discrete_terminating_event_occurred = (
+        "Terminating solve because a discrete event occurred."
+    )
     max_steps_reached = (
         "The maximum number of solver steps was reached. Try increasing `max_steps`."
     )
     dt_min_reached = "The minimum step size was reached."
-    nan_time = "NaN time encountered during timestepping."
     implicit_divergence = "Implicit method diverged."
     implicit_nonconvergence = (
         "Implicit method did not converge within the required number of iterations."
     )
-    discrete_terminating_event_occurred = (
-        "Terminating solve because a discrete event occurred."
-    )
 
 
 def is_okay(result: RESULTS) -> Bool:
@@ -86,8 +85,8 @@ class Solution(AbstractPath):
         must allocate enough space for the maximum possible number of steps.
     """
 
-    t0: Scalar = field(init=True)
-    t1: Scalar = field(init=True)  # override init=False in AbstractPath
+    t0: Scalar = field(init=True, repr=True)
+    t1: Scalar = field(init=True, repr=True)  # override AbstractPath
     ts: Optional[Array["times"]]  # noqa: F821
     ys: Optional[PyTree["times", ...]]  # noqa: F821
     interpolation: Optional[DenseInterpolation]
