@@ -4,7 +4,7 @@ import diffrax
 import jax.numpy as jnp
 import pytest
 
-from helpers import shaped_allclose
+from .helpers import shaped_allclose
 
 
 def test_results():
